# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.saving namespace.
"""

import sys as _sys

from keras.saving.object_registration import CustomObjectScope as custom_object_scope
from keras.saving.object_registration import get_custom_objects
from keras.saving.object_registration import get_registered_name
from keras.saving.object_registration import get_registered_object
from keras.saving.object_registration import register_keras_serializable
from keras.saving.saving_api import load_model
from keras.saving.saving_api import save_model