# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""VGG19 model for Keras.

Reference:
  - [Very Deep Convolutional Networks for Large-Scale Image Recognition](
      https://arxiv.org/abs/1409.1556) (ICLR 2015)

"""

import sys as _sys

from keras.applications.vgg19 import VGG19
from keras.applications.vgg19 import decode_predictions
from keras.applications.vgg19 import preprocess_input