# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.__internal__.models namespace.
"""

import sys as _sys

from keras.models.cloning import clone_and_build_model
from keras.models.cloning import in_place_subclassed_model_state_restoration